import pytest
from graia.saya import Saya
from graia.saya.behaviour.entity import Behaviour
from graia.saya.schema import BaseSchema

from launart.manager import Launart
from launart.saya import LaunartBehaviour


class EmptyBehaviour(Behaviour):
    def allocate(self, cube):
        if cube.metaclass.__class__ is BaseSchema:
            return True

    def release(self, cube):
        if cube.metaclass.__class__ is BaseSchema:
            return True


def test_saya():
    saya = Saya()
    mgr = Launart()
    saya.install_behaviours(LaunartBehaviour(mgr), EmptyBehaviour())

    empty_mod = saya.require("tests._saya_mod.empty_sub")

    with pytest.raises(TypeError):
        saya.require("tests._saya_mod.fail_sub")

    channel = saya.require("tests._saya_mod.ok_sub")
    assert "launchable.test.saya" in mgr.launchables
    assert "service.test.saya" in mgr.launchables
    saya.uninstall_channel(channel)
    assert "launchable.test.saya" not in mgr.launchables
    assert "service.test.saya" not in mgr.launchables

    saya.uninstall_channel(empty_mod)



from graia.saya.channel import Channel

from launart.component import Launchable
from launart.saya import LaunchableSchema
from launart.service import ExportInterface, Service

c = Channel.current()


SayaTestInterface = type("SayaTestInterface", (ExportInterface,), {})


class SayaTestLaunchable(Launchable):
    id = "launchable.test.saya"

    @property
    def required(self):
        return set()

    @property
    def stages(self):
        return set()

    async def launch(self, _):
        ...


class SayaTestSrv(Service):
    supported_interface_types = {SayaTestInterface}
    id = "service.test.saya"

    @property
    def required(self):
        return set()

    @property
    def stages(self):
        return set()

    async def launch(self, _):
        ...

    def get_interface(self, interface_type):
        return interface_type()


c.use(LaunchableSchema())(SayaTestLaunchable())
c.use(LaunchableSchema())(SayaTestSrv())



import asyncio

import pytest

from launart import Launart
from launart.component import Launchable, LaunchableStatus
from tests.fixture import EmptyLaunchable, component, interface, service


def test_ensure():
    lc = EmptyLaunchable()
    mgr = Launart()
    mgr.add_launchable(lc)
    assert lc.manager is mgr
    with pytest.raises(RuntimeError):
        lc.ensure_manager(Launart())


def test_resolve_req():
    i = interface()
    s = service("srv", {i}, [])
    a = component("a", ["b", i])
    b = component("b", [])
    mgr = Launart()
    with pytest.raises(RuntimeError):
        a._required_id
    mgr.add_launchable(a)
    with pytest.raises(RuntimeError):
        a._required_id
    mgr.add_launchable(s)
    with pytest.raises(RuntimeError):
        a._required_id
    mgr.add_launchable(b)
    assert a._required_id == {"srv", "b"}


def test_launchable_stat_transition_raw():
    stat = LaunchableStatus()
    stat.stage = "blocking"
    assert stat.stage == "blocking"
    with pytest.raises(ValueError):
        stat.stage = "preparing"  # rollback is not allowed
    stat.stage = "blocking-completed"
    stat.stage = "finished"
    stat.unset()


def test_launchable_stat_transition_base_err_report():
    class _Base(Launchable):
        @property
        def required(self):
            return set()

        @property
        def stages(self):
            return {"preparing"}

    class ErrNoMgr(_Base):
        id = "e1"

        async def launch(self, _):
            async with self.stage("preparing"):
                ...

    with pytest.raises(RuntimeError):
        asyncio.run(ErrNoMgr().launch(None))
    with pytest.raises(LookupError):
        mgr = Launart()
        e = ErrNoMgr()
        assert not e.manager
        e.ensure_manager(mgr)
        asyncio.run(e.launch(None))

    class ErrUnexpectedStage(_Base):
        id = "e2"

        async def launch(self, _):
            async with self.stage("blocking"):  # oops
                ...

    with pytest.raises(ValueError):
        mgr = Launart()
        e = ErrUnexpectedStage()
        e.ensure_manager(mgr)
        mgr.status.stage = "preparing"
        asyncio.run(e.launch(None))

    class ErrUnknownStageDef(_Base):
        id = "e2"

        @property
        def stages(self):
            return {"finished"}

        async def launch(self, _):
            async with self.stage("finished"):  # oops
                ...

    with pytest.raises(ValueError):
        mgr = Launart()
        e = ErrUnknownStageDef()
        e.ensure_manager(mgr)
        mgr.status.stage = "preparing"
        asyncio.run(e.launch(None))


def test_launchable_stat_transition_err():
    class _Base(Launchable):
        @property
        def required(self):
            return set()

        @property
        def stages(self):
            return {"preparing", "blocking", "cleanup"}

    class ErrToPrepare(_Base):
        id = "e1"

        async def launch(self, _):
            async with self.stage("preparing"):
                ...
            async with self.stage("blocking"):
                ...
            async with self.stage("cleanup"):
                ...
            with pytest.raises(ValueError):
                async with self.stage("preparing"):
                    ...

    class ErrToBlocking(_Base):
        id = "e2"

        async def launch(self, _):
            async with self.stage("preparing"):
                ...
            async with self.stage("blocking"):
                ...
            async with self.stage("cleanup"):
                ...
            with pytest.raises(ValueError):
                async with self.stage("blocking"):
                    ...

    class ErrToCleanup(_Base):
        id = "e3"

        async def launch(self, _):
            async with self.stage("preparing"):
                ...
            async with self.stage("blocking"):
                ...
            async with self.stage("cleanup"):
                ...
            with pytest.raises(ValueError):
                async with self.stage("cleanup"):
                    ...

    mgr = Launart()

    for e in [ErrToPrepare(), ErrToBlocking(), ErrToCleanup()]:
        mgr.add_launchable(e)
    mgr.launch_blocking()



import asyncio
from typing import Any, cast

import pytest

from launart._sideload import Override, override
from launart.component import RequirementResolveFailed, resolve_requirements
from launart.utilles import priority_strategy, wait_fut
from tests.fixture import component_standalone, interface


def test_resolve_success():
    dataset = [
        component_standalone("a", []),
        component_standalone("b", ["a"]),
        component_standalone("c", ["a", "b"]),
        component_standalone("d", ["a", "c"]),
        component_standalone("e", []),
        component_standalone("f", ["a", "c"]),
    ]
    expected = [{dataset[0], dataset[4]}, {dataset[1]}, {dataset[2]}, {dataset[3], dataset[5]}]
    assert resolve_requirements(dataset) == expected
    expected.reverse()
    assert resolve_requirements(dataset, reverse=True) == expected


def test_resolve_fail():
    dataset = [
        component_standalone("a", ["b"]),
        component_standalone("b", ["a"]),
    ]
    with pytest.raises(RequirementResolveFailed):
        resolve_requirements(dataset)


def test_priority_strategy():
    def spit(v: Any) -> Any:
        return v

    with pytest.raises(TypeError):
        priority_strategy([1, 2, 3], spit)
    with pytest.raises(TypeError):
        priority_strategy([(1, 2), 3], spit)

    i = [interface() for _ in range(10)]

    s1 = {i[0], i[1]}
    s2 = {i[1], i[2]}
    s3 = {i[2], i[3]}

    d1 = {i[0]: 1, i[1]: 2}
    d2 = {i[1]: 1, i[2]: 2}
    d3 = {i[2]: 1, i[3]: 2}

    assert priority_strategy([s1, s3], spit) == {i[0]: s1, i[1]: s1, i[2]: s3, i[3]: s3}

    assert priority_strategy([d1, d2, d3], spit) == {i[0]: d1, i[1]: d1, i[2]: d2, i[3]: d3}

    assert priority_strategy([d1, d3, d2], spit) == {i[0]: d1, i[1]: d1, i[2]: d2, i[3]: d3}

    assert priority_strategy([(d1, d2), d3], spit) == {i[0]: (d1, d2), i[1]: (d1, d2), i[2]: (d1, d2), i[3]: d3}

    with pytest.raises(ValueError):
        priority_strategy([s1, s2], spit)
    with pytest.raises(ValueError):
        priority_strategy([s1, d1], spit)


def test_override():
    class MyOrigin:
        data: dict

        def __init__(self, d: dict) -> None:
            self.data = d

    origin = MyOrigin({"a": 3})
    additional = {"a": 4}
    o = cast(Override, override(origin, additional))
    assert o.source is origin
    assert o.a == 4
    with pytest.raises(AttributeError):
        o.b


@pytest.mark.asyncio
async def test_wait_fut():
    await wait_fut([])
    await wait_fut([asyncio.sleep(0.01), asyncio.create_task(asyncio.sleep(0.02))])

    t1 = asyncio.create_task(asyncio.sleep(0.01))
    t2 = asyncio.create_task(asyncio.sleep(0.1))

    await wait_fut([t1, t2], timeout=0.02)

    assert t1.done()
    assert not t2.done()

    t2.cancel()

    t1 = asyncio.create_task(asyncio.sleep(0.01))
    t2 = asyncio.create_task(asyncio.sleep(0.1))

    await wait_fut([t1, t2], return_when=asyncio.FIRST_COMPLETED)

    assert t1.done()
    assert not t2.done()

    t2.cancel()



from __future__ import annotations

import asyncio
import contextlib
import signal
from contextvars import ContextVar
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Iterable,
    Literal,
    Optional,
    Type,
    TypeVar,
    cast,
)

from loguru import logger
from statv import Stats, Statv

from launart._sideload import FutureMark, override
from launart.component import Launchable, resolve_requirements
from launart.service import ExportInterface, Service, TInterface
from launart.utilles import FlexibleTaskGroup, priority_strategy

U_ManagerStage = Literal["preparing", "blocking", "cleaning", "finished"]
E = TypeVar("E", bound=ExportInterface)


def _launchable_task_done_callback(mgr: "Launart", t: asyncio.Task):  # pragma: no cover
    try:
        exc = t.exception()
    except asyncio.CancelledError:
        logger.warning(
            f"[{t.get_name()}] was cancelled in abort.",
            alt=f"[yellow bold]component [magenta]{t.get_name()}[/] was cancelled in abort.",
        )
        return
    if exc:
        logger.opt(exception=exc).error(
            f"[{t.get_name()}] raised a exception.",
            alt=f"[red bold]component [magenta]{t.get_name()}[/] raised an exception.",
        )
        return

    component = mgr.get_launchable(t.get_name())
    if mgr.status.preparing:
        if "preparing" in component.stages:
            if component.status.prepared:
                logger.info(f"Component {t.get_name()} completed preparation.")
            else:
                logger.error(f"Component {t.get_name()} exited before preparation.")
    elif mgr.status.blocking:
        if "cleanup" in component.stages and component.status.stage != "finished":
            logger.warning(f"Component {t.get_name()} exited without cleanup.")
        else:
            logger.success(f"Component {t.get_name()} finished.")
    elif mgr.status.cleaning:
        if "cleanup" in component.stages:
            if component.status.finished:
                logger.success(f"component {t.get_name()} finished.")
            else:
                logger.warning(f"component {t.get_name()} exited before completing cleanup.")

    logger.info(
        f"[{t.get_name()}] completed.",
        alt=rf"[green]\[[magenta]{t.get_name()}[/magenta]] completed.",
    )
    if not component.status.finished:
        component.status.stage = "finished"


class ManagerStatus(Statv):
    stage = Stats[Optional[U_ManagerStage]]("U_ManagerStage", default=None)
    exiting = Stats[bool]("exiting", default=False)

    def __init__(self) -> None:
        super().__init__()

    def __repr__(self) -> str:
        return f"<ManagerStatus stage={self.stage} waiters={len(self._waiters)}>"

    @property
    def preparing(self) -> bool:
        return self.stage == "preparing"

    @property
    def blocking(self) -> bool:
        return self.stage == "blocking"

    @property
    def cleaning(self) -> bool:
        return self.stage == "cleaning"

    async def wait_for_update(self, *, current: str | None = None, stage: U_ManagerStage | None = None):
        waiter = asyncio.Future()
        if current is not None:
            waiter.add_done_callback(FutureMark(current, stage))
        self._waiters.append(waiter)
        try:
            return await waiter
        finally:
            self._waiters.remove(waiter)

    async def wait_for_preparing(self):
        while not self.preparing:
            await self.wait_for_update()

    async def wait_for_blocking(self):
        while not self.blocking:
            await self.wait_for_update()

    async def wait_for_cleaning(self, *, current: str | None = None):
        while not self.cleaning:
            await self.wait_for_update(current=current, stage="cleaning")

    async def wait_for_finished(self, *, current: str | None = None):
        while self.stage not in {"finished", None}:
            await self.wait_for_update(current=current, stage="finished")

    async def wait_for_sigexit(self):
        while self.stage in {"preparing", "blocking"} and not self.exiting:
            await self.wait_for_update()


class Launart:
    launchables: Dict[str, Launchable]
    status: ManagerStatus
    tasks: dict[str, asyncio.Task]
    task_group: Optional[FlexibleTaskGroup] = None

    _service_bind: Dict[Type[ExportInterface], Service]

    _context: ClassVar[ContextVar[Launart]] = ContextVar("launart._context")
    _priority_overrides: Dict[Type[ExportInterface], Service]

    def __init__(self):
        self.launchables = {}
        self._service_bind = {}
        self.tasks = {}
        self.status = ManagerStatus()
        self._priority_overrides = {}

    @classmethod
    def current(cls) -> Launart:
        return cls._context.get()

    def add_launchable(self, launchable: Launchable):
        launchable.ensure_manager(self)
        if launchable.id in self.launchables:
            raise ValueError(f"Launchable {launchable.id} already exists.")
        if self.task_group is not None:
            launchable._required_id  # Evaluate requirements automatically.
            tracker = asyncio.create_task(self._sideload_tracker(launchable))
            self.task_group.sideload_trackers[launchable.id] = tracker
            self.task_group.add(tracker)  # flush the waiter tasks
        self.launchables[launchable.id] = launchable
        if isinstance(launchable, Service):
            self._update_service_bind()

    def get_launchable(self, id: str) -> Launchable:
        if id not in self.launchables:
            raise ValueError(f"Launchable {id} does not exists.")
        return self.launchables[id]

    def get_service(self, id: str) -> Service:
        launchable = self.get_launchable(id)
        if not isinstance(launchable, Service):
            raise TypeError(f"{id} is not a service.")
        return launchable

    def override_bind(self, interface: type[ExportInterface], service: Service) -> None:
        if interface in self._priority_overrides:
            raise ValueError(f"{interface} is already overridden by {self._priority_overrides[interface]}")
        self._priority_overrides[interface] = service
        self._service_bind.update(self._priority_overrides)

    def remove_launchable(
        self,
        launchable: str | Launchable,
    ):
        if isinstance(launchable, str):
            if launchable not in self.launchables:
                if self.task_group and launchable in self.task_group.sideload_trackers:
                    return
                raise ValueError(f"Launchable {id} does not exist.")
            target = self.launchables[launchable]
        else:
            target = launchable
        if self.task_group is None:
            del self.launchables[target.id]
        else:
            if target.id not in self.task_group.sideload_trackers:
                raise RuntimeError("Only sideload tasks can be removed at runtime!")
            tracker = self.task_group.sideload_trackers[target.id]
            if tracker.cancelled() or tracker.done():  # completed in silence, let it pass
                return
            if target.status.stage not in {"prepared", "blocking", "blocking-completed", "waiting-for-cleanup"}:
                raise RuntimeError(
                    f"{target.id} obtains invalid stats to sideload active release, it's {target.status.stage}"
                )
            tracker.cancel()  # trigger cancel, and the tracker will start clean up the
        if isinstance(target, Service):
            self._update_service_bind()

    def _update_service_bind(self):
        self._service_bind = priority_strategy(
            [i for i in self.launchables.values() if isinstance(i, Service)],
            lambda a: a.supported_interface_types,
        )
        self._service_bind.update(self._priority_overrides)

    add_service = add_launchable
    remove_service = remove_launchable

    def get_interface(self, interface_type: Type[TInterface]) -> TInterface:
        if interface_type not in self._service_bind:
            raise ValueError(f"{interface_type} is not supported.")
        return self._service_bind[interface_type].get_interface(interface_type)

    def _get_task(self, launchable_id: str):
        if self.task_group is None:
            raise RuntimeError("Task Group is not initialized.")
        for key, task in self.tasks.items():
            if key == launchable_id:
                return task

    async def _sideload_tracker(self, launchable: Launchable) -> None:
        if TYPE_CHECKING:
            assert self.task_group is not None

        logger.info(f"Sideload {launchable.id}: injecting")

        local_status = ManagerStatus()
        shallow_self = cast(Launart, override(self, {"status": local_status}))
        launchable.manager = shallow_self

        task = asyncio.create_task(launchable.launch(shallow_self), name=launchable.id)
        task.add_done_callback(partial(_launchable_task_done_callback, self))
        self.tasks[launchable.id] = task

        local_status.stage = "preparing"
        if "preparing" in launchable.stages:
            await self._sideload_prepare(launchable)

        with contextlib.suppress(asyncio.CancelledError):
            local_status.stage = "blocking"
            if "blocking" in launchable.stages:
                await self._sideload_blocking(launchable)

        local_status.update_multi(
            {
                ManagerStatus.stage: "cleaning",
                ManagerStatus.exiting: True,
            }
        )
        if "cleanup" in launchable.stages:
            await self._sideload_cleanup(launchable)

        if not task.done() or not task.cancelled():  # pragma: worst case
            await task
        logger.info(f"Sideload {launchable.id}: completed.")
        del self.tasks[launchable.id]
        del self.launchables[launchable.id]

    async def _sideload_prepare(self, launchable: Launchable) -> None:
        if launchable.status.stage != "waiting-for-prepare":  # pragma: worst case
            logger.info(f"waiting sideload {launchable.id} for prepare")
            await asyncio.wait(
                [
                    self.tasks[launchable.id],
                    asyncio.create_task(launchable.status.wait_for("waiting-for-prepare")),
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )

        logger.info(f"Sideload {launchable.id}: preparing")

        launchable.status.stage = "preparing"
        await asyncio.wait(
            [
                self.tasks[launchable.id],
                asyncio.create_task(launchable.status.wait_for("prepared")),
            ],
            return_when=asyncio.FIRST_COMPLETED,
        )
        logger.info(f"Sideload {launchable.id}: preparation completed")

    async def _sideload_blocking(self, launchable: Launchable) -> None:

        logger.info(f"Sideload {launchable.id}: start blocking")

        await asyncio.wait(
            [
                self.tasks[launchable.id],
                asyncio.create_task(launchable.status.wait_for("blocking-completed")),
            ],
            return_when=asyncio.FIRST_COMPLETED,
        )
        logger.info(f"Sideload {launchable.id}: blocking completed")

    async def _sideload_cleanup(self, launchable: Launchable):

        if launchable.status.stage != "waiting-for-cleanup":  # pragma: worst case
            await asyncio.wait(
                [
                    self.tasks[launchable.id],
                    asyncio.create_task(launchable.status.wait_for("waiting-for-cleanup")),
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )

        launchable.status.stage = "cleanup"

        await asyncio.wait(
            [
                self.tasks[launchable.id],
                asyncio.create_task(launchable.status.wait_for("finished")),
            ],
            return_when=asyncio.FIRST_COMPLETED,
        )
        logger.info(f"Sideload {launchable.id}: cleanup completed.")

    async def launch(self):  # sourcery skip: low-code-quality
        _token = self._context.set(self)
        if self.status.stage is not None:
            logger.error("Incorrect ownership, launart is already running.")
            return
        self.status.exiting = False
        logger.info(
            f"Launching {len(self.launchables)} components as async task...",
            alt=f"Launching [magenta]{len(self.launchables)}[/] components as async task...",
        )

        loop = asyncio.get_running_loop()
        as_task = loop.create_task
        self.tasks = {}

        for launchable in self.launchables.values():
            task = loop.create_task(launchable.launch(self), name=launchable.id)
            task.add_done_callback(partial(_launchable_task_done_callback, self))
            self.tasks[launchable.id] = task

        self.status.stage = "preparing"
        for k, v in self.launchables.items():
            if "preparing" in v.stages and v.status.stage != "waiting-for-prepare":
                await asyncio.wait(
                    [self.tasks[k], as_task(v.status.wait_for("waiting-for-prepare"))],
                    return_when=asyncio.FIRST_COMPLETED,
                )
        for layer, components in enumerate(
            resolve_requirements(self.launchables.values()),
            start=1,
        ):
            preparing_tasks = []
            tracking_tasks: list[asyncio.Task] = []
            for i in components:
                if "preparing" in i.stages:
                    i.status.stage = "preparing"
                    tracking_tasks.append(self.tasks[i.id])
                    preparing_tasks.append(
                        asyncio.wait(
                            [self.tasks[i.id], as_task(i.status.wait_for("prepared"))],
                            return_when=asyncio.FIRST_COMPLETED,
                        )
                    )
            if preparing_tasks:
                await asyncio.gather(*preparing_tasks)

                if any(tsk.done() and tsk.exception() is not None for tsk in tracking_tasks):
                    msg = f"Layer #{layer}:[{', '.join(tsk.get_name() for tsk in tracking_tasks if tsk.exception() is not None)}] failed during preparation. Aborting."
                    logger.critical(
                        msg,
                        alt=f"[green]Layer[/] [magenta]#{layer}[/]:"
                        f"[red bold][{', '.join(f'[cyan italic]{tsk.get_name()}[/cyan italic]' for tsk in tracking_tasks if tsk.exception() is not None)}] failed during preparation. Aborting.[/]",
                    )
                    for tsk in self.tasks.values():
                        tsk.cancel()
                    self.status.stage = None
                    return

                logger.success(
                    f"Layer #{layer}:[{', '.join([i.id for i in components])}] preparation completed.",
                    alt=f"[green]Layer [magenta]#{layer}[/]:[{', '.join([f'[cyan italic]{i.id}[/cyan italic]' for i in components])}] preparation completed.",
                )

        logger.info("All components prepared, start blocking phase.", style="green bold")

        self.status.stage = "blocking"
        self.task_group = FlexibleTaskGroup()
        blocking_tasks = [
            asyncio.wait(
                [
                    self.tasks[i.id],
                    as_task(i.status.wait_for("blocking-completed")),
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )
            for i in self.launchables.values()
            if "blocking" in i.stages
        ]
        self.task_group.add(*blocking_tasks)
        try:
            if blocking_tasks:
                await self.task_group
        except asyncio.CancelledError:
            logger.info("Blocking phase cancelled by user.", style="red bold")
        finally:
            self.status.exiting = True
            logger.info("Entering cleanup phase.", style="yellow bold")
            # cleanup the dangling sideload tasks first.
            if self.task_group.sideload_trackers:
                for tracker in self.task_group.sideload_trackers.values():
                    tracker.cancel()
                await asyncio.wait(self.task_group.sideload_trackers.values())
            self.status.stage = "cleaning"
            for k, v in self.launchables.items():
                if "cleanup" in v.stages and v.status.stage != "waiting-for-cleanup":
                    await asyncio.wait(
                        [
                            self.tasks[k],
                            as_task(v.status.wait_for("waiting-for-cleanup")),
                        ],
                        return_when=asyncio.FIRST_COMPLETED,
                    )
            for layer, components in enumerate(
                resolve_requirements(self.launchables.values(), reverse=True),
                start=1,
            ):
                cleaning_tasks = []
                for i in components:
                    if "cleanup" in i.stages and i.status.stage == "waiting-for-cleanup":
                        i.status.stage = "cleanup"
                        cleaning_tasks.append(
                            asyncio.wait(
                                [
                                    self.tasks[i.id],
                                    as_task(i.status.wait_for("finished")),
                                ],
                                return_when=asyncio.FIRST_COMPLETED,
                            )
                        )
                if cleaning_tasks:
                    await asyncio.gather(*cleaning_tasks)

                    logger.success(
                        f"Layer #{layer}:[{', '.join([i.id for i in components])}] cleanup completed.",
                        alt=f"[green]Layer [magenta]#{layer}[/]:[{', '.join([f'[cyan]{i.id}[/cyan]' for i in components])}] cleanup completed.",
                    )

            self.status.stage = "finished"
            logger.info(
                "Cleanup completed, waiting for finalization.",
                style="green bold",
            )

            finale_tasks = [i for i in self.tasks.values() if not i.done()]
            if finale_tasks:
                await asyncio.wait(finale_tasks)
            self.task_group = None
            self._context.reset(_token)
            logger.success("All launch task finished.", style="green bold")
            self.status.stage = None  # reset status

    def launch_blocking(
        self,
        *,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        stop_signal: Iterable[signal.Signals] = (signal.SIGINT,),
    ):
        import contextlib
        import functools
        import threading

        loop = loop or asyncio.new_event_loop()

        logger.info("Starting launart main task...", style="green bold")

        launch_task = loop.create_task(self.launch(), name="amnesia-launch")
        handled_signals: Dict[signal.Signals, Any] = {}
        signal_handler = functools.partial(self._on_sys_signal, main_task=launch_task)
        if threading.current_thread() is threading.main_thread():  # pragma: worst case
            try:
                for sig in stop_signal:
                    handled_signals[sig] = signal.getsignal(sig)
                    signal.signal(sig, signal_handler)
            except ValueError:  # pragma: no cover
                # `signal.signal` may throw if `threading.main_thread` does
                # not support signals
                handled_signals.clear()

        loop.run_until_complete(launch_task)

        for sig, handler in handled_signals.items():
            if signal.getsignal(sig) is signal_handler:
                signal.signal(sig, handler)

        try:
            self._cancel_tasks(loop)
            loop.run_until_complete(loop.shutdown_asyncgens())
            with contextlib.suppress(RuntimeError, AttributeError):
                # LINK: https://docs.python.org/3.10/library/asyncio-eventloop.html#asyncio.loop.shutdown_default_executor
                loop.run_until_complete(loop.shutdown_default_executor())
        finally:
            logger.success("asyncio shutdown complete.", style="green bold")

    def _on_sys_signal(self, _, __, main_task: asyncio.Task):
        self.status.exiting = True
        if self.task_group is not None:
            self.task_group.stop = True
            if self.task_group.blocking_task is not None:  # pragma: worst case
                self.task_group.blocking_task.cancel()
        if not main_task.done():
            main_task.cancel()
            # wakeup loop if it is blocked by select() with long timeout
            main_task._loop.call_soon_threadsafe(lambda: None)
            logger.warning("Ctrl-C triggered by user.", style="dark_orange bold")

    @staticmethod
    def _cancel_tasks(loop: asyncio.AbstractEventLoop):
        import asyncio
        import asyncio.tasks

        to_cancel = asyncio.tasks.all_tasks(loop)
        if to_cancel:
            for tsk in to_cancel:
                tsk.cancel()
            loop.run_until_complete(asyncio.gather(*to_cancel, return_exceptions=True))

            for task in to_cancel:  # pragma: no cover
                # BELIEVE IN PSF
                if task.cancelled():
                    continue
                if task.exception() is not None:
                    logger.opt(exception=task.exception()).error(f"Unhandled exception when shutting down {task}:")



import asyncio
from signal import SIGINT, default_int_handler, signal

import pytest

from launart import Launart
from launart.component import Launchable
from launart.service import ExportInterface, Service
from tests.fixture import EmptyLaunchable, component, interface, service


@pytest.mark.asyncio
async def test_nothing():
    mgr = Launart()
    lc = EmptyLaunchable()
    mgr.add_launchable(lc)
    await mgr.launch()
    assert lc.triggered


def test_nothing_blocking():
    mgr = Launart()
    lc = EmptyLaunchable()
    mgr.add_launchable(lc)
    mgr.launch_blocking()
    assert lc.triggered

    # set SIGINT and do again
    lc.triggered = False
    loop = asyncio.new_event_loop()
    tsk = loop.create_task(asyncio.sleep(2.0))
    tsk2 = loop.create_task(asyncio.sleep(0))
    signal(SIGINT, lambda *_: None)
    mgr.launch_blocking(loop=loop)
    assert tsk.cancelled()
    assert tsk.done()
    assert tsk2.done()
    assert lc.triggered
    signal(SIGINT, default_int_handler)


def test_nothing_complex():
    mgr = Launart()

    class _L(Launchable):
        id = "empty"
        triggered = False

        @property
        def stages(self):
            return {"blocking"}

        @property
        def required(self):
            return set()

        async def launch(self, _):
            async with self.stage("blocking"):
                await asyncio.sleep(0.2)

    lc = _L()
    mgr.add_launchable(lc)

    loop = asyncio.new_event_loop()
    launch_tsk = loop.create_task(mgr.launch())
    loop.run_until_complete(asyncio.sleep(0.02))  # head time
    assert mgr._get_task("empty")
    wrong_launch = loop.create_task(mgr.launch())
    loop.run_until_complete(asyncio.sleep(0.1))
    mgr._on_sys_signal(None, None, launch_tsk)
    assert not launch_tsk.done()
    assert wrong_launch.done()
    loop.run_until_complete(launch_tsk)
    mgr._on_sys_signal(None, None, launch_tsk)


def test_manager_stat():
    mgr = Launart()

    class _L(Launchable):
        id = "test_stat"

        @property
        def stages(self):
            return {"preparing", "blocking", "cleanup"}

        @property
        def required(self):
            return set()

        async def launch(self, _):
            await self.status.wait_for()  # test empty wait
            await asyncio.sleep(0.02)
            async with self.stage("preparing"):
                await asyncio.sleep(0.02)
            await asyncio.sleep(0.02)
            assert self.status.prepared
            async with self.stage("blocking"):
                assert self.status.prepared
                assert self.status.blocking
                await asyncio.sleep(0.02)
            await asyncio.sleep(0.02)
            async with self.stage("cleanup"):
                assert not self.status.prepared
                await asyncio.sleep(0.02)
            await asyncio.sleep(0.02)

    mgr.add_launchable(_L())
    loop = asyncio.new_event_loop()
    mk_task = loop.create_task
    tasks = [mk_task(mgr.status.wait_for_preparing())]
    loop.run_until_complete(asyncio.sleep(0.01))
    mk_task(mgr.launch())
    tasks.append(mk_task(mgr.status.wait_for_blocking()))
    tasks.append(mk_task(mgr.status.wait_for_cleaning()))
    loop.run_until_complete(asyncio.sleep(0.02))
    tasks.append(mk_task(mgr.status.wait_for_sigexit()))
    tasks.append(mk_task(mgr.status.wait_for_finished()))
    loop.run_until_complete(asyncio.sleep(0.2))
    for task in tasks:
        assert task.done() and not task.cancelled()


@pytest.mark.asyncio
async def test_wait_for(event_loop: asyncio.AbstractEventLoop):
    loop = event_loop
    mgr = Launart()

    class _L(Launchable):
        id = "test_stat"

        @property
        def stages(self):
            return {"preparing", "blocking", "cleanup"}

        @property
        def required(self):
            return set()

        async def launch(self, _):
            await self.status.wait_for()  # test empty wait
            await asyncio.sleep(0.02)
            async with self.stage("preparing"):
                await asyncio.sleep(0.02)
            await asyncio.sleep(0.02)
            assert self.status.prepared
            async with self.stage("blocking"):
                assert self.status.prepared
                assert self.status.blocking
                await asyncio.sleep(0.02)
            await asyncio.sleep(0.02)
            async with self.stage("cleanup"):
                assert not self.status.prepared
                await asyncio.sleep(0.02)

    l = _L()
    with pytest.raises(RuntimeError):
        await l.wait_for("preparing", "test_stat")
    mgr.add_launchable(l)
    mk_task = loop.create_task
    mk_task(mgr.launch())
    await l.wait_for("finished", "test_stat")


def test_signal_change_during_running():
    mgr = Launart()

    class _L(Launchable):
        id = "empty"
        triggered = False

        @property
        def stages(self):
            return {"blocking"}

        @property
        def required(self):
            return set()

        async def launch(self, _):
            signal(SIGINT, lambda *_: None)

    lc = _L()
    mgr.add_launchable(lc)

    mgr.launch_blocking()


def test_bare_bone():
    mgr = Launart()
    lc = component("component.test", [])
    mgr.add_launchable(lc)
    assert mgr.launchables["component.test"] == lc
    i = interface()
    srv = service("service.test", {i}, [])
    mgr.add_launchable(srv)
    assert mgr.launchables["service.test"] == srv
    assert mgr.get_launchable("service.test") is srv
    assert mgr.get_service("service.test") is srv
    assert isinstance(mgr.get_interface(i), i)
    with pytest.raises(ValueError):
        mgr.get_interface(interface())
    with pytest.raises(TypeError):
        mgr.get_service("component.test")
    with pytest.raises(ValueError):
        mgr.add_launchable(lc)
    with pytest.raises(ValueError):
        mgr.get_launchable("$?")


@pytest.mark.asyncio
async def test_basic_components():
    TestInterface = type("SayaTestInterface", (ExportInterface,), {})

    stage = []

    class TestLaunchable(Launchable):
        id = "launchable.test"

        @property
        def required(self):
            return {TestInterface}

        @property
        def stages(self):
            return {"preparing", "blocking", "cleanup"}

        async def launch(self, _):
            async with self.stage("preparing"):
                stage.append("lc prepare")
            async with self.stage("blocking"):
                stage.append("blocking")
                assert isinstance(mgr._get_task("service.test"), asyncio.Task)
                assert mgr._get_task("nothing") is None
            async with self.stage("cleanup"):
                stage.append("lc cleanup")

    class TestSrv(Service):
        supported_interface_types = {TestInterface}
        id = "service.test"

        @property
        def required(self):
            return set()

        @property
        def stages(self):
            return {"preparing", "blocking", "cleanup"}

        async def launch(self, _):
            async with self.stage("preparing"):
                stage.append("srv prepare")
            async with self.stage("blocking"):
                stage.append("blocking")
            async with self.stage("cleanup"):
                stage.append("srv cleanup")

        def get_interface(self, interface_type):
            return interface_type()

    mgr = Launart()
    mgr.add_launchable(TestLaunchable())
    mgr.add_launchable(TestSrv())
    with pytest.raises(RuntimeError):
        mgr._get_task("service.test")
    await mgr.launch()
    assert stage == ["srv prepare", "lc prepare", "blocking", "blocking", "lc cleanup", "srv cleanup"]


def test_override_bind():
    TestInterface = type("SayaTestInterface", (ExportInterface,), {})

    class Srv1(Service):
        supported_interface_types = {TestInterface: 3}
        id = "service.test1"

        @property
        def required(self):
            return set()

        @property
        def stages(self):
            return {"preparing", "blocking", "cleanup"}

        async def launch(self, _):
            ...

        def get_interface(self, interface_type):
            return interface_type()

    class Srv2(Service):
        supported_interface_types = {TestInterface: 1}
        id = "service.test2"

        @property
        def required(self):
            return set()

        @property
        def stages(self):
            return {"preparing", "blocking", "cleanup"}

        async def launch(self, _):
            ...

        def get_interface(self, interface_type):
            return interface_type()

    mgr = Launart()
    srv1, srv2 = Srv1(), Srv2()
    mgr.add_launchable(srv1)
    assert mgr._service_bind[TestInterface] is srv1
    mgr.add_launchable(srv2)
    assert mgr._service_bind[TestInterface] is srv1
    mgr.override_bind(TestInterface, srv2)
    assert mgr._service_bind[TestInterface] is srv2
    with pytest.raises(ValueError):
        mgr.override_bind(TestInterface, srv1)


def test_graceful_abort():

    failure: bool = False

    class Malfunction(Launchable):
        id = "malfunction"

        @property
        def required(self):
            return set()

        @property
        def stages(self):
            return {"preparing"}

        async def launch(self, _):
            async with self.stage("preparing"):
                raise ValueError

    class Dependent(Launchable):
        id = "dependent"

        @property
        def required(self):
            return {"malfunction"}

        @property
        def stages(self):
            return {"preparing", "blocking"}

        async def launch(self, _):
            async with self.stage("preparing"):
                ...
            async with self.stage("blocking"):
                nonlocal failure
                failure = True

    mgr = Launart()
    mgr.add_launchable(Malfunction())
    mgr.add_launchable(Dependent())
    mgr.launch_blocking()

    if failure:
        pytest.fail("Error: dependent reached blocking stage while dependency failed to prepare.")



from __future__ import annotations

import asyncio
import enum
from typing import (
    Callable,
    Coroutine,
    Dict,
    Hashable,
    Iterable,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar("T")
H = TypeVar("H", bound=Hashable)


class _Unmarked(enum.Enum):
    UNMARKED = object()


UNMARKED = _Unmarked.UNMARKED

# "Unmarked" is a great way of replacing ellipsis.

PriorityType = Union[
    Set[T],
    Dict[T, float],
    Tuple[
        Union[
            Set[T],
            Dict[T, float],
        ],
        ...,
    ],
]


def priority_strategy(
    items: List[T],
    getter: Callable[
        [T],
        PriorityType[H],
    ],
) -> Dict[H, T]:
    result: Dict[H, T] = {}
    _priority_mem: Dict[H, float | Literal[UNMARKED]] = {}

    def _handle(pattern: PriorityType[H]) -> None:
        """Handle an actual pattern."""
        if isinstance(pattern, set):
            # Pattern has unknown priorities.
            for content in pattern:
                if content in _priority_mem:
                    raise ValueError(f"{content} conflicts with {result[content]}")
                _priority_mem[content] = UNMARKED
                result[content] = item

        elif isinstance(pattern, dict):
            for content, priority in pattern.items():
                if content in _priority_mem:
                    current_priority: float | Literal[_Unmarked.UNMARKED] = _priority_mem[content]
                    if current_priority is UNMARKED or priority is UNMARKED:
                        raise ValueError(f"Unable to determine priority order: {content}, {result[content]}.")
                    if current_priority < priority:
                        _priority_mem[content] = priority
                        result[content] = item
                else:
                    _priority_mem[content] = priority
                    result[content] = item

        else:
            raise TypeError(f"{pattern} is not a valid pattern.")

    for item in items:
        pattern: PriorityType[H] = getter(item)
        if isinstance(pattern, (dict, set)):
            _handle(pattern)
        elif isinstance(pattern, tuple):
            for sub_pattern in pattern:
                _handle(sub_pattern)
        else:
            raise TypeError(f"{pattern} is not a valid pattern.")
    return result


async def wait_fut(
    coros: Iterable[Union[Coroutine, asyncio.Task]],
    *,
    timeout: Optional[float] = None,
    return_when: str = asyncio.ALL_COMPLETED,
) -> None:
    tasks = []
    for c in coros:
        if asyncio.iscoroutine(c):
            tasks.append(asyncio.create_task(c))
        else:
            tasks.append(c)
    if tasks:
        await asyncio.wait(tasks, timeout=timeout, return_when=return_when)


class FlexibleTaskGroup:
    tasks: list[asyncio.Task]
    sideload_trackers: dict[str, asyncio.Task]
    blocking_task: Optional[asyncio.Task] = None
    stop: bool = False

    def __init__(self, *tasks):
        self.sideload_trackers = {}
        self.tasks = list(tasks)

    def __await__(self):
        return self.__await_impl__().__await__()

    async def __await_impl__(self):
        while True:
            self.blocking_task = asyncio.create_task(asyncio.wait(self.tasks))
            try:
                return await self.blocking_task
            except asyncio.CancelledError:
                if self.stop:
                    raise

    def add(self, *fs: asyncio.Task | Coroutine) -> None:
        tasks = [f if isinstance(f, asyncio.Task) else asyncio.create_task(f) for f in fs]
        if self.blocking_task is not None:
            self.blocking_task.cancel()
        self.tasks.extend(tasks)



from __future__ import annotations

from collections import ChainMap
from typing import TYPE_CHECKING, Any, Dict, TypeVar, cast

if TYPE_CHECKING:
    from launart.manager import U_ManagerStage


class Override:
    def __init__(self, source: Any, additional: Dict[str, Any]):
        self.__source = source
        self.__additional = additional

    @property
    def source(self):
        return self.__source

    @property
    def __data(self):
        return ChainMap(self.__additional, vars(self.__source))

    def __getattr__(self, item):
        if item not in self.__data:
            raise AttributeError(f"'{self.__source.__class__.__name__}' object has no attribute '{item}'")
        return self.__data[item]


T = TypeVar("T")


def override(source: T, additional: Dict[str, Any]) -> T:
    return cast(T, Override(source, additional))


class FutureMark:
    id: str
    stage: U_ManagerStage | None = None

    def __init__(self, id: str, stage: U_ManagerStage | None = None) -> None:
        self.id = id
        self.stage = stage

    def __call__(self, _) -> Any:
        pass



import asyncio

import pytest

from launart import Launart
from launart.component import Launchable
from tests.fixture import component


class _Base(Launchable):
    @property
    def required(self):
        return set()

    @property
    def stages(self):
        return {"preparing", "blocking", "cleanup"}


@pytest.mark.asyncio
async def test_sideload_base():
    s1 = asyncio.Event()
    s2 = asyncio.Event()
    s_cleanup = asyncio.Event()

    class Sideload1(_Base):
        id = "worker1"

        async def launch(self, manager: Launart):
            async with self.stage("preparing"):
                ...
            async with self.stage("blocking"):
                s1.set()
                await s_cleanup.wait()
            async with self.stage("cleanup"):
                pass

    class Sideload2(_Base):
        id = "worker2"

        async def launch(self, manager: Launart):
            async with self.stage("preparing"):
                ...
            async with self.stage("blocking"):
                s2.set()
                await s_cleanup.wait()
            async with self.stage("cleanup"):
                ...

    class AddSideload(_Base):
        id = "test_sideload_add"

        async def launch(self, manager: Launart):
            async with self.stage("preparing"):
                ...
            async with self.stage("blocking"):
                manager.add_launchable(Sideload1())
                manager.add_launchable(component("worker_none", []))
                with pytest.raises(RuntimeError):
                    manager.add_launchable(component("worker_weird_req", ["$"]))
                with pytest.raises(RuntimeError):
                    manager.remove_launchable("worker1")  # wrong status
                with pytest.raises(RuntimeError):
                    manager.remove_launchable("test_sideload_add")  # removing prohibited
                await s1.wait()
                manager.add_launchable(Sideload2())
                await s2.wait()
                manager.remove_launchable("worker1")
                s_cleanup.set()
            async with self.stage("cleanup"):
                with pytest.raises(ValueError):
                    manager.remove_launchable("worker-any")
                manager.remove_launchable("worker2")

    mgr = Launart()
    mgr.add_launchable(AddSideload())
    await mgr.launch()



from __future__ import annotations

from launart import Launart, Launchable
from launart.service import ExportInterface, Service
from launart.utilles import PriorityType


class EmptyLaunchable(Launchable):
    id = "empty"
    triggered = False

    @property
    def stages(self):
        return set()

    @property
    def required(self):
        return set()

    async def launch(self, manager: Launart):
        assert Launart.current() is manager
        assert manager is self.manager
        from loguru import logger

        logger.success("EmptyLaunchable Triggered")
        self.triggered = True


def component_standalone(component_id: str, required_ids: list[str]) -> Launchable:
    class _L(Launchable):
        id = component_id

        @property
        def required(self):
            return set(required_ids)

        @property
        def _required_id(self):
            return set(required_ids)

        @property
        def stages(self):
            return set()

        async def launch(self, _):
            ...

    return _L()


def component(component_id: str, required_ids: list[str | type[ExportInterface]]) -> Launchable:
    class _L(Launchable):
        id = component_id

        @property
        def required(self):
            return set(required_ids)

        @property
        def stages(self):
            return set()

        async def launch(self, _):
            ...

    return _L()


def interface() -> type[ExportInterface]:
    return type("$NewInterface", (ExportInterface,), {})


def service(component_id: str, interfaces: PriorityType, required_ids: list[str | type[ExportInterface]]) -> Service:
    class Srv(Service):
        supported_interface_types = interfaces
        id = component_id

        @property
        def required(self):
            return set(required_ids)

        @property
        def stages(self):
            return set()

        async def launch(self, _):
            ...

        def get_interface(self, interface_type):
            return interface_type()

    return Srv()



from __future__ import annotations

from dataclasses import dataclass

from graia.saya.behaviour import Behaviour
from graia.saya.cube import Cube
from graia.saya.schema import BaseSchema

from launart import Launart, Launchable


@dataclass
class LaunchableSchema(BaseSchema):
    pass


class LaunartBehaviour(Behaviour):
    manager: Launart

    def __init__(self, manager: Launart) -> None:
        self.manager = manager

    def allocate(self, cube: Cube[LaunchableSchema]):
        if isinstance(cube.metaclass, LaunchableSchema):
            if not isinstance(cube.content, Launchable):
                raise TypeError(f"{cube.content} is not a Launchable")
            self.manager.add_launchable(cube.content)
        else:
            return
        return True

    def release(self, cube: Cube[LaunchableSchema]):
        if isinstance(cube.metaclass, LaunchableSchema):
            self.manager.remove_launchable(cube.content)
        else:
            return
        return True

    uninstall = release



from contextlib import suppress as _suppress

from .component import Launchable as Launchable
from .component import LaunchableStatus as LaunchableStatus
from .component import RequirementResolveFailed as RequirementResolveFailed
from .component import U_Stage as U_Stage
from .manager import Launart as Launart
from .service import ExportInterface as ExportInterface
from .service import Service as Service

with _suppress(ImportError, ModuleNotFoundError):
    from .saya import LaunartBehaviour as LaunartBehaviour
    from .saya import LaunchableSchema as LaunchableSchema
del _suppress



from __future__ import annotations

import asyncio
from abc import ABCMeta, abstractmethod
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Iterable, List, Optional, Set, Union

from statv import Stats, Statv

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

if TYPE_CHECKING:
    from launart.manager import Launart
    from launart.service import ExportInterface

U_Stage = Union[
    Literal[
        "waiting-for-prepare",
        "preparing",
        "prepared",
        "blocking",
        "blocking-completed",
        "waiting-for-cleanup",
        "cleanup",
        "finished",
    ],
    None,
]
STAGE_STAT = {
    None: {"waiting-for-prepare", "waiting-for-cleanup", "blocking", "finished"},
    "waiting-for-prepare": {"preparing"},
    "preparing": {"prepared"},
    "prepared": {"blocking", "waiting-for-cleanup", "finished"},
    "blocking": {"blocking-completed"},
    "blocking-completed": {"waiting-for-cleanup", "finished"},
    "waiting-for-cleanup": {"cleanup"},
    "cleanup": {"finished"},
    "finished": {None},
}
# for runtime check:
STATS = [
    None,
    "waiting-for-prepare",
    "preparing",
    "prepared",
    "blocking",
    "blocking-completed",
    "waiting-for-cleanup",
    "cleanup",
    "finished",
]


class LaunchableStatus(Statv):
    stage = Stats[Optional[U_Stage]]("stage", default=None)

    def __init__(self) -> None:
        super().__init__()

    @property
    def prepared(self) -> bool:
        return self.stage in ("prepared", "blocking")

    @property
    def blocking(self) -> bool:
        return self.stage == "blocking"

    @property
    def finished(self) -> bool:
        return self.stage == "finished"

    @staticmethod
    @stage.validator
    def _(stats: Stats[U_Stage | None], past: U_Stage | None, current: U_Stage | None):
        if current not in STAGE_STAT[past]:
            raise ValueError(f"Invalid stage transition: {past} -> {current}")
        return current

    def unset(self) -> None:
        self.stage = None

    async def wait_for(self, stage: U_Stage = None):
        stages = set(STATS[STATS.index(stage) :])
        while self.stage not in stages:
            await self.wait_for_update()


class Launchable(metaclass=ABCMeta):
    id: str
    status: LaunchableStatus
    manager: Optional[Launart] = None

    def __init__(self) -> None:
        self.status = LaunchableStatus()

    @property
    @abstractmethod
    def required(self) -> Set[str | type[ExportInterface]]:
        ...

    @property
    def _required_id(self) -> Set[str]:
        """Extract ID of combined requirements."""
        dependencies: Set[str] = set()
        failed: Set[str | type[ExportInterface]] = set()
        if self.manager is None:
            raise RuntimeError("attempted to flatten dependencies without a manager.")
        for req in self.required:
            if req not in (self.manager.launchables if isinstance(req, str) else self.manager._service_bind):
                failed.add(req)
            elif isinstance(req, str):
                dependencies.add(req)
            else:
                dependencies.add(self.manager._service_bind[req].id)
        if failed:
            raise RuntimeError(f"Failed to find following requirements: {failed!r}")
        return dependencies

    @property
    @abstractmethod
    def stages(self) -> Set[Literal["preparing", "blocking", "cleanup"]]:
        ...

    def ensure_manager(self, manager: Launart):
        if self.manager is not None and self.manager is not manager:
            raise RuntimeError("this launchable attempted to be mistaken a wrong ownership of launart/manager.")
        self.manager = manager

    @asynccontextmanager
    async def stage(self, stage: U_Stage):
        if self.manager is None:
            raise RuntimeError("attempted to set stage of a launchable without a manager.")
        if self.manager.status.stage is None:
            raise LookupError("attempted to set stage of a launchable without a current manager")
        if stage not in self.stages:
            raise ValueError(f"undefined and unexpected stage entering: {stage}")

        if stage == "preparing":
            if "waiting-for-prepare" not in STAGE_STAT[self.status.stage]:
                raise ValueError(f"unexpected stage entering: {self.status.stage} -> waiting-for-prepare")
            await self.manager.status.wait_for_preparing()
            self.status.stage = "waiting-for-prepare"
            await self.status.wait_for("preparing")
            yield
            self.status.stage = "prepared"
        elif stage == "blocking":
            if "blocking" not in STAGE_STAT[self.status.stage]:
                raise ValueError(f"unexpected stage entering: {self.status.stage} -> blocking")
            await self.manager.status.wait_for_blocking()
            await self.wait_for_required()
            self.status.stage = "blocking"
            yield
            self.status.stage = "blocking-completed"
        elif stage == "cleanup":
            if "waiting-for-cleanup" not in STAGE_STAT[self.status.stage]:
                raise ValueError(f"unexpected stage entering: {self.status.stage} -> waiting-for-cleanup")
            await self.manager.status.wait_for_cleaning(current=self.id)
            self.status.stage = "waiting-for-cleanup"
            await self.status.wait_for("cleanup")
            yield
            self.status.stage = "finished"
        else:
            raise ValueError(f"entering unexpected stage: {stage}(unknown definition)")

    async def wait_for_required(self, stage: U_Stage = "prepared"):
        await self.wait_for(stage, *self._required_id)

    async def wait_for(self, stage: U_Stage, *launchable_id: str):
        if self.manager is None:
            raise RuntimeError("attempted to set stage of a launchable without a manager.")
        launchables = [self.manager.get_launchable(id) for id in launchable_id]
        while any(launchable.status.stage not in STATS[STATS.index(stage) :] for launchable in launchables):
            await asyncio.wait(
                [
                    asyncio.create_task(launchable.status.wait_for_update())
                    for launchable in launchables
                    if launchable.status.stage != stage
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )

    @abstractmethod
    async def launch(self, manager: Launart):
        pass


class RequirementResolveFailed(ValueError):
    pass


def resolve_requirements(components: Iterable[Launchable], reverse: bool = False) -> List[Set[Launchable]]:
    resolved_id: Set[str] = set()
    unresolved: Set[Launchable] = set(components)
    result: List[Set[Launchable]] = []
    while unresolved:
        layer = {component for component in unresolved if resolved_id >= component._required_id}

        if layer:
            unresolved -= layer
            resolved_id.update(component.id for component in layer)
            result.append(layer)
        else:
            raise RequirementResolveFailed(unresolved)
    if reverse:
        result.reverse()
    return result
